module Bin where

import Word
import IO
import Bits
import Time
import Int
import GenUtil
import Char

type BinHandle = Handle

putBits :: BinHandle -> Int -> Word32 -> IO ()
putBits _ n _ | n < 0 || n > 32 = error $ "putBits on invalid count: " ++ show n
putBits _ 0 _ = return ()
putBits bh nbits val = do
    hPutChar bh ("01" !! ((fromIntegral val) `mod` 2))
    putBits bh (nbits - 1) (val `shiftR` 1)
getBits :: BinHandle -> Int -> IO Word32
getBits _ nbits | nbits < 0 || nbits > 32 = error $ "getBits on invalid count: " ++ show nbits
getBits _ 0 = return 0
getBits bh nbits = do
    xs <- replicateM nbits (hGetChar bh)
    let cs = map f $ reverse xs 
    return $ foldl (\c x -> (c `shiftL` 1) + x ) 0 cs where
	f '0' = 0
	f '1' = 1
    
class Binary a where
    put    :: BinHandle -> a -> IO ()
    get    :: BinHandle -> IO a
    sizeOf :: a -> Int

instance Binary () where
    put  bh () = return ()
    get  _     = return ()
    sizeOf _ = 0

instance Binary Bool where
    put  bh b = putBits bh 1 (fromIntegral $ fromEnum b)
    get  bh   = getBits bh 1 >>= return . toEnum . fromIntegral
    sizeOf _ = 1

instance Binary Char where
    put  bh c = put bh (ord c)
    get  bh   = do x <- get bh; return (chr x)
    sizeOf _ = sizeOf (undefined :: Int)

instance Binary Int where
    put  bh i = putBits bh 32 (fromIntegral i)
    get  bh   = fmap fromIntegral $ getBits bh 32
    sizeOf _ = 32

instance Binary Word8 where
    put  bh i = putBits bh 8 (fromIntegral i)
    get  bh   = fmap fromIntegral $  getBits bh 8
    sizeOf _ = 8

instance Binary Int32 where
    put  bh i = putBits bh 32 (fromIntegral i)
    get  bh   = fmap fromIntegral $  getBits bh 32
    sizeOf _ = 32


instance Binary ClockTime where
    put bh ct = do
	let t = toUTCTime ct
	put bh (ctYear t)
	put bh (fromEnum $ ctMonth t)
	put bh (ctDay t)
	put bh (ctHour t)
	put bh (ctMin t)
	put bh (ctSec t)
    get bh = do
	year <- get bh
	month <- fmap toEnum $ get bh 
	day <- get bh 
	hour <- get bh 
	min <- get bh 
	sec <- get bh 
	return $ toClockTime $ (toUTCTime epoch) {ctYear = year, ctDay = day, ctMonth = month, ctHour = hour, ctMin = min, ctSec = sec}
    sizeOf _ = 32*6
	

instance (Binary a, Binary b) => Binary (a,b) where
    put bh (a,b) = do put bh a
                      put bh b
    get bh       = do a <- get bh
                      b <- get bh
                      return (a,b)
    sizeOf (a,b) = sizeOf a + sizeOf b

instance Binary a => Binary [a] where
    put bh []     = putBits bh 1 0
    put bh (x:xs) = do putBits bh 1 1
                       put bh x
                       put bh xs
    get bh        = do h <- getBits bh 1
                       case h of
                          0 -> return []
                          _ -> do x  <- get bh
                                  xs <- get bh
                                  return (x:xs)
    sizeOf xs = sum (map sizeOf xs) + length xs + 1

